
#include "ll_ifc.h"
#include "ll_ifc_lorawan.h"
#include "ll_ifc_no_mac.h"
#include "embc/fifo.h"
#include "embc/log.h"
#include <stdint.h>
#include <string.h> // memset, memcpy

#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>
#include "cmockery.h"

#define TRANSPORT_LEN_MAX       256
#define TRANSPORT_INVOKE_MAX    16


struct fifo8_s * rd_fifo;

uint32_t log_timestamp_us( void ) {
    return 0;
}

void log_printf(const char * format, ...) {
    (void) format;
}

void dbc_assert(char const *file, unsigned line, const char * msg) {
    (void) file;
    (void) line;
    (void) msg;
}

int32_t gettime(struct time *tp)
{
    if (tp) {
        tp->tv_sec = 0;
        tp->tv_nsec = 0;
    }
    return 0;
}

int32_t sleep_ms(int32_t millis)
{
    (void) millis;
    return 0;
}

struct transport_single_s {
    uint8_t wr_buf[TRANSPORT_LEN_MAX];
    uint16_t wr_len;
    uint8_t rd_buf[TRANSPORT_LEN_MAX];
    uint16_t rd_len;
};

struct transport_expect_s {
    struct transport_single_s buffer[TRANSPORT_INVOKE_MAX];
    int head;
    int tail;

    uint8_t message_num;
    uint8_t msg_buf[TRANSPORT_LEN_MAX];
    uint16_t msg_len;
};

struct transport_expect_s transport_expect_;

static uint16_t compute_checksum(uint8_t const * buf, uint16_t len)
{
    uint16_t i;
    uint16_t crc = 0;
    for(i = 0; i < len; i++)
    {
        crc = (crc >> 8) | (crc << 8);
        crc ^= buf[i];
        crc ^= (crc & 0xff) >> 4;
        crc ^= crc << 12;
        crc ^= (crc & 0xff) << 5;
    }
    return crc;
}

void transport_expect(
        uint8_t opcode,
        uint8_t const * wr_buf, uint16_t wr_len,
        uint8_t ack,
        uint8_t const * rd_buf, uint16_t rd_len)
{
    uint16_t crc;
    struct transport_expect_s * self = &transport_expect_;
    struct transport_single_s * s;
    int next = self->head + 1;
    if (next >= TRANSPORT_INVOKE_MAX)
    {
        next = 0;
    }
    assert_true(next != self->tail);
    assert_true(wr_len <= TRANSPORT_LEN_MAX - 11);
    assert_true(rd_len <= TRANSPORT_LEN_MAX - 8);
    s = &self->buffer[self->head];

    // construct command
    s->wr_len = 11 + wr_len;
    s->wr_buf[0] = 0xff;
    s->wr_buf[1] = 0xff;
    s->wr_buf[2] = 0xff;
    s->wr_buf[3] = 0xff;
    s->wr_buf[4] = 0xc4;
    s->wr_buf[5] = opcode;
    s->wr_buf[6] = self->message_num;
    s->wr_buf[7] = (wr_len >> 8) & 0xff;
    s->wr_buf[8] = (wr_len >> 0) & 0xff;
    memcpy(&s->wr_buf[9], wr_buf, wr_len);
    crc = compute_checksum(s->wr_buf + 4, s->wr_len - 6);
    s->wr_buf[s->wr_len - 2] = (crc >> 8) & 0xff;
    s->wr_buf[s->wr_len - 1] = (crc >> 0) & 0xff;

    // construct response
    s->rd_len = 8 + rd_len;
    s->rd_buf[0] = 0xc4;
    s->rd_buf[1] = opcode;
    s->rd_buf[2] = self->message_num;
    s->rd_buf[3] = ack;
    s->rd_buf[4] = (rd_len >> 8) & 0xff;
    s->rd_buf[5] = (rd_len >> 0) & 0xff;
    memcpy(&s->rd_buf[6], rd_buf, rd_len);
    crc = compute_checksum(s->rd_buf, s->rd_len - 2);
    s->rd_buf[s->rd_len - 2] = (crc >> 8) & 0xff;
    s->rd_buf[s->rd_len - 1] = (crc >> 0) & 0xff;

    self->head = next;
    self->message_num++;
}

int32_t transport_write(uint8_t *buff, uint16_t len)
{
    struct transport_expect_s * self = &transport_expect_;
    struct transport_single_s * s;

    assert_true(self->head != self->tail); // "write when none expected");
    assert_true(len + self->msg_len <= TRANSPORT_LEN_MAX); // "receive buffer overflow");
    memcpy(&self->msg_buf[self->msg_len], buff, len);
    self->msg_len += len;

    s = &self->buffer[self->tail];
    if (self->msg_len >= s->wr_len)
    {
        int next = self->tail + 1;
        if (next >= TRANSPORT_INVOKE_MAX)
        {
            next = 0;
        }
        assert_memory_equal(s->wr_buf, self->msg_buf, s->wr_len);
        memcpy(self->msg_buf, self->msg_buf + s->wr_len, self->msg_len - s->wr_len);
        self->msg_len -= s->wr_len;
        if (s->rd_len) {
            fifo_size_t actual_size;
            assert_int_equal(0, fifo8_write(rd_fifo, s->rd_buf, (fifo_size_t) s->rd_len, &actual_size));
            assert_int_equal(actual_size, s->rd_len);
        }
        self->tail = next;
    }
    return 0;
}

int32_t transport_read(uint8_t *buff, uint16_t len)
{
    uint16_t i;
    for (i = 0; i < len; ++i) {
        if (fifo8_read_u8(rd_fifo, buff++)) {
            return 1;
        }
    }
    return 0;
}

void setUp(void **state)
{
    (void) state;
    memset(&transport_expect_, 0, sizeof(transport_expect_));
    assert_int_equal(0, fifo8_alloc(&rd_fifo, TRANSPORT_LEN_MAX + TRANSPORT_INVOKE_MAX));
    assert_true(0 != rd_fifo);
    ll_reset_state();
}

void tearDown(void **state)
{
    (void) state;
    fifo8_free(rd_fifo);
}

void test_myself_transport_write(void **state)
{
    (void) state;
    uint8_t cmd_payload[] = {1, 2, 3, 4};
    uint8_t cmd[] = {0xff, 0xff, 0xff, 0xff, 0xc4, 1, 0, 0, 4, 1, 2, 3, 4, 0, 0};
    uint8_t rsp_payload[] = {5, 6, 7, 8};
    uint8_t rsp[TRANSPORT_LEN_MAX];

    uint16_t crc = compute_checksum(cmd + 4, sizeof(cmd) - 6);
    cmd[sizeof(cmd) - 2] = (crc >> 8) & 0xff;
    cmd[sizeof(cmd) - 1] = (crc >> 0) & 0xff;

    transport_expect(1, cmd_payload, sizeof(cmd_payload), 0, rsp_payload, sizeof(rsp_payload));
    assert_int_equal(0, transport_write(cmd, sizeof(cmd)));
    assert_int_equal(0, transport_read(rsp, sizeof(rsp_payload) + 8));
    assert_memory_equal(rsp_payload, rsp + 6, sizeof(rsp_payload));
}

void test_timestamp_get(void **state)
{
    (void) state;
    uint8_t cmd[] = {0x00, 0x00, 0x00, 0x00, 0x00};
    uint8_t rsp[] = {0x11, 0x22, 0x33, 0x44};

    transport_expect(OP_TIMESTAMP, cmd, sizeof(cmd), 0, rsp, sizeof(rsp));

    uint32_t actual_timestamp_us = 0;
    assert_int_equal(0, ll_timestamp_get(&actual_timestamp_us));
    assert_int_equal(0x11223344, actual_timestamp_us);
}

void test_timestamp_get_with_response_error(void **state)
{
    (void) state;
    uint8_t cmd[] = {0x00, 0x00, 0x00, 0x00, 0x00};
    uint8_t rsp[] = {0x11, 0x22, 0x33}; // too short

    transport_expect(OP_TIMESTAMP, cmd, sizeof(cmd), 0, rsp, sizeof(rsp));

    uint32_t actual_timestamp_us = 0;
    assert_int_equal(LL_IFC_ERROR_INCORRECT_RESPONSE_LENGTH, ll_timestamp_get(&actual_timestamp_us));
}

void test_timestamp_get_twice_to_check_message_identifier(void **state)
{
    (void) state;
    uint8_t cmd[] = {0x00, 0x00, 0x00, 0x00, 0x00};
    uint8_t rsp[] = {0x11, 0x22, 0x33, 0x44};

    transport_expect(OP_TIMESTAMP, cmd, sizeof(cmd), 0, rsp, sizeof(rsp));
    transport_expect(OP_TIMESTAMP, cmd, sizeof(cmd), 0, rsp, sizeof(rsp));

    uint32_t actual_timestamp_us = 0;
    assert_int_equal(0, ll_timestamp_get(&actual_timestamp_us));
    assert_int_equal(0x11223344, actual_timestamp_us);

    assert_int_equal(0, ll_timestamp_get(&actual_timestamp_us));
    assert_int_equal(0x11223344, actual_timestamp_us);
}

void test_timestamp_set(void **state)
{
    (void) state;
    uint8_t cmd[] = {0x01, 0x10, 0x20, 0x30, 0x40};
    uint8_t rsp[] = {0x11, 0x22, 0x33, 0x44};

    transport_expect(OP_TIMESTAMP, cmd, sizeof(cmd), 0, rsp, sizeof(rsp));

    uint32_t actual_timestamp_us = 0;
    assert_int_equal(0, ll_timestamp_set(1, 0x10203040, &actual_timestamp_us));
    assert_int_equal(0x11223344, actual_timestamp_us);
}

void test_timestamp_set_with_nack(void **state)
{
    (void) state;
    uint8_t cmd[] = {0x01, 0x10, 0x20, 0x30, 0x40};

    transport_expect(OP_TIMESTAMP, cmd, sizeof(cmd), LL_IFC_NACK_BUSY_TRY_AGAIN, 0, 0);

    uint32_t actual_timestamp_us = 0;
    assert_int_equal(-LL_IFC_NACK_BUSY_TRY_AGAIN, ll_timestamp_set(1, 0x10203040, &actual_timestamp_us));
}

void test_nack_with_payload(void **state)
{
    (void) state;
    uint8_t cmd[] = {0x01, 0x10, 0x20, 0x30, 0x40};
    uint8_t rsp[] = {1};

    transport_expect(OP_TIMESTAMP, cmd, sizeof(cmd), LL_IFC_NACK_BUSY_TRY_AGAIN, rsp, sizeof(rsp));

    uint32_t actual_timestamp_us = 0;
    // todo : indicating a CRC error does not seem correct here
    assert_int_equal(-104, ll_timestamp_set(1, 0x10203040, &actual_timestamp_us));
}

void test_timestamp_set_with_invalid_operation(void **state)
{
    (void) state;
    uint32_t actual_timestamp_us = 0;
    assert_int_equal(LL_IFC_ERROR_INCORRECT_PARAMETER,
            ll_timestamp_set((ll_timestamp_operation_t) -1,
                             0x10203040, &actual_timestamp_us));
    assert_int_equal(LL_IFC_ERROR_INCORRECT_PARAMETER,
            ll_timestamp_set((ll_timestamp_operation_t) LL_TIMESTAMP_SYNC + 1,
                             0x10203040, &actual_timestamp_us));
}

void test_packet_send_timestamp(void **state)
{
    (void) state;
    uint8_t cmd1[] = {0x10, 0x20, 0x30, 0x40};

    uint8_t cmd2[] = {1, 2, 3, 4, 5, 6, 7, 8};
    uint8_t rsp2[] = {0};

    transport_expect(OP_SEND_TIMESTAMP, cmd1, sizeof(cmd1), 0, 0, 0);
    transport_expect(OP_PKT_SEND_QUEUE, cmd2, sizeof(cmd2), 0, rsp2, sizeof(rsp2));
    assert_int_equal(0, ll_packet_send_timestamp(0x10203040, cmd2, sizeof(cmd2)));
}


/* ----------------------- LoRaWAN ----------------------------------------- */

void test_lorawan_activate_over_air(void **state)
{
    (void) state;

    uint8_t devEui[] = {0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17};
    uint8_t appEui[] = {0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27};
    uint8_t appKey[] = {0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
                        0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f};

    uint8_t cmd1[] = {0, 0, 0,
            0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
            0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
            0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
            0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f
    };
    uint8_t rsp_success[] = {LL_LORAWAN_ACTIVATION_STATUS_COMPLETED};
    uint8_t rsp_failed[] = {LL_LORAWAN_ACTIVATION_STATUS_FAILED};
    uint8_t rsp_pending[] = {LL_LORAWAN_ACTIVATION_STATUS_PENDING};

    uint8_t cmd_wait[] = {LL_LORAWAN_ACTIVATION_QUERY};
    uint8_t cmd_irq[] = {0x00, 0x00, 0x11, 0x30};
    uint8_t rsp_irq[] = {0, 0, 0, 0};

    // bad network
    assert_int_equal(LL_IFC_ERROR_INCORRECT_PARAMETER,
            ll_lorawan_activate_over_the_air((enum ll_lorawan_network_type_e) 99,
                    LL_LORAWAN_CLASS_A, devEui, appEui, appKey));

    // IFC protocol error
    transport_expect(OP_LORAWAN_ACTIVATE, cmd1, sizeof(cmd1), LL_IFC_NACK_OTHER, 0, 0);
    assert_int_equal(-LL_IFC_NACK_OTHER, ll_lorawan_activate_over_the_air(LL_LORAWAN_PUBLIC, LL_LORAWAN_CLASS_A, devEui, appEui, appKey));

    // immediate success (unexpected, but ok)
    transport_expect(OP_LORAWAN_ACTIVATE, cmd1, sizeof(cmd1), 0, rsp_success, sizeof(rsp_success));
    assert_int_equal(0, ll_lorawan_activate_over_the_air(LL_LORAWAN_PUBLIC, LL_LORAWAN_CLASS_A, devEui, appEui, appKey));

    // immediate failure (unexpected, but failure)
    transport_expect(OP_LORAWAN_ACTIVATE, cmd1, sizeof(cmd1), 0, rsp_failed, sizeof(rsp_failed));
    assert_int_equal(-LL_IFC_NACK_OTHER, ll_lorawan_activate_over_the_air(LL_LORAWAN_PUBLIC, LL_LORAWAN_CLASS_A, devEui, appEui, appKey));

    // pending and then succeed
    transport_expect(OP_LORAWAN_ACTIVATE, cmd1, sizeof(cmd1), 0, rsp_pending, sizeof(rsp_pending));
    transport_expect(OP_LORAWAN_ACTIVATE, cmd_wait, sizeof(cmd_wait), 0, rsp_pending, sizeof(rsp_pending));
    transport_expect(OP_LORAWAN_ACTIVATE, cmd_wait, sizeof(cmd_wait), 0, rsp_success, sizeof(rsp_success));
    transport_expect(OP_IRQ_FLAGS, cmd_irq, sizeof(cmd_irq), 0, rsp_irq, sizeof(rsp_irq));
    assert_int_equal(0, ll_lorawan_activate_over_the_air(LL_LORAWAN_PUBLIC, LL_LORAWAN_CLASS_A, devEui, appEui, appKey));

    // pending and then fail
    transport_expect(OP_LORAWAN_ACTIVATE, cmd1, sizeof(cmd1), 0, rsp_pending, sizeof(rsp_pending));
    transport_expect(OP_LORAWAN_ACTIVATE, cmd_wait, sizeof(cmd_wait), 0, rsp_pending, sizeof(rsp_pending));
    transport_expect(OP_LORAWAN_ACTIVATE, cmd_wait, sizeof(cmd_wait), 0, rsp_failed, sizeof(rsp_failed));
    assert_int_equal(-LL_IFC_NACK_OTHER, ll_lorawan_activate_over_the_air(LL_LORAWAN_PUBLIC, LL_LORAWAN_CLASS_A, devEui, appEui, appKey));
}

void test_lorawan_activate_personalization(void **state)
{
    (void) state;

    uint32_t netID = 0x00112233;
    uint32_t devAddr = 0x55667788;
    uint8_t netSKey[] = {0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
                         0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f};
    uint8_t appSKey[] = {0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
                         0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f};

    uint8_t cmd1[] = {1, 0, 0,
            0x00, 0x11, 0x22, 0x33,
            0x55, 0x66, 0x77, 0x88,
            0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
            0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
            0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
            0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f
    };
    uint8_t rsp_success[] = {LL_LORAWAN_ACTIVATION_STATUS_COMPLETED};
    uint8_t rsp_pending[] = {LL_LORAWAN_ACTIVATION_STATUS_PENDING};

    uint8_t cmd_wait[] = {LL_LORAWAN_ACTIVATION_QUERY};
    uint8_t cmd_irq[] = {0x00, 0x00, 0x11, 0x30};
    uint8_t rsp_irq[] = {0, 0, 0, 0};

    // bad network
    assert_int_equal(LL_IFC_ERROR_INCORRECT_PARAMETER,
            ll_lorawan_activate_personalization((enum ll_lorawan_network_type_e) 99,
                    LL_LORAWAN_CLASS_A, netID, devAddr, netSKey, appSKey));

    // pending and then succeed
    transport_expect(OP_LORAWAN_ACTIVATE, cmd1, sizeof(cmd1), 0, rsp_pending, sizeof(rsp_pending));
    transport_expect(OP_LORAWAN_ACTIVATE, cmd_wait, sizeof(cmd_wait), 0, rsp_pending, sizeof(rsp_pending));
    transport_expect(OP_LORAWAN_ACTIVATE, cmd_wait, sizeof(cmd_wait), 0, rsp_success, sizeof(rsp_success));
    transport_expect(OP_IRQ_FLAGS, cmd_irq, sizeof(cmd_irq), 0, rsp_irq, sizeof(rsp_irq));
    assert_int_equal(0, ll_lorawan_activate_personalization(LL_LORAWAN_PUBLIC, LL_LORAWAN_CLASS_A, netID, devAddr, netSKey, appSKey));
}

void test_lorawan_param_get_i32_invalid(void **state)
{
    (void) state;

    int32_t value = 0;
    uint8_t cmd1[] = {3, LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 0};

    assert_int_equal(LL_IFC_ERROR_INCORRECT_PARAMETER,
            ll_lorawan_param_get_i32((enum ll_lorawan_param_e) 1000, &value));

    transport_expect(OP_LORAWAN_PARAM, cmd1, sizeof(cmd1), LL_IFC_NACK_OTHER, 0, 0);
    assert_int_equal(-LL_IFC_NACK_OTHER, ll_lorawan_param_get_i32(LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, &value));
}

void test_lorawan_param_get_i32_success(void **state)
{
    (void) state;

    int32_t value = 0;
    uint8_t cmd1[] = {3, LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 0};
    uint8_t rsp1[] = {3, LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 4, 0x11, 0x22, 0x33, 0x44};

    transport_expect(OP_LORAWAN_PARAM, cmd1, sizeof(cmd1), 0, rsp1, sizeof(rsp1));
    assert_int_equal(0, ll_lorawan_param_get_i32(LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, &value));
    assert_int_equal(0x11223344, value);
}

void test_lorawan_param_get_i32_with_payload_too_short(void **state)
{
    (void) state;

    int32_t value = 0;
    uint8_t cmd1[] = {3, LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 0};
    uint8_t rsp1[] = {3, LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 4, 0x11, 0x22};

    transport_expect(OP_LORAWAN_PARAM, cmd1, sizeof(cmd1), 0, rsp1, sizeof(rsp1));
    assert_int_equal(LL_IFC_ERROR_INCORRECT_MESSAGE_SIZE, ll_lorawan_param_get_i32(LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, &value));
}

void test_lorawan_param_set_i32_invalid(void **state)
{
    (void) state;

    uint8_t cmd1[] = {3, LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 4, 0x11, 0x22, 0x33, 0x44};

    assert_int_equal(LL_IFC_ERROR_INCORRECT_PARAMETER,
            ll_lorawan_param_set_i32((enum ll_lorawan_param_e) 1000, 0x11223344));

    transport_expect(OP_LORAWAN_PARAM, cmd1, sizeof(cmd1), LL_IFC_NACK_OTHER, 0, 0);
    assert_int_equal(-LL_IFC_NACK_OTHER, ll_lorawan_param_set_i32(LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 0x11223344));
}

void test_lorawan_param_set_i32_success(void **state)
{
    (void) state;

    uint8_t cmd1[] = {3, LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 4, 0x11, 0x22, 0x33, 0x44};
    uint8_t rsp1[] = {3, LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 4, 0x11, 0x22, 0x33, 0x44};

    transport_expect(OP_LORAWAN_PARAM, cmd1, sizeof(cmd1), 0, rsp1, sizeof(rsp1));
    assert_int_equal(0, ll_lorawan_param_set_i32(LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 0x11223344));
}

void test_lorawan_param_set_i32_with_payload_too_short(void **state)
{
    (void) state;

    uint8_t cmd1[] = {3, LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 4, 0x11, 0x22, 0x33, 0x44};
    uint8_t rsp1[] = {3, LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 4, 0x11, 0x22};

    transport_expect(OP_LORAWAN_PARAM, cmd1, sizeof(cmd1), 0, rsp1, sizeof(rsp1));
    assert_int_equal(LL_IFC_ERROR_INCORRECT_MESSAGE_SIZE, ll_lorawan_param_set_i32(LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, 0x11223344));
}

void test_lorawan_send_packet_unconfirmed(void **state)
{
    (void) state;
    uint8_t msg[] = {0x11, 0x22, 0x33, 0x44, 0x55};
    uint8_t cmd1[] = {LL_LORAWAN_SEND_LINK_CHECK,
            7, 0, 5, 0x11, 0x22, 0x33, 0x44, 0x55};
    uint8_t rsp_success[] = {0};

    transport_expect(OP_LORAWAN_MSG_SEND, cmd1, sizeof(cmd1), 0, rsp_success, sizeof(rsp_success));
    assert_int_equal(0,
            ll_lorawan_send_unconfirmed(LL_LORAWAN_SEND_LINK_CHECK, 7, msg, sizeof(msg)));
}

void test_lorawan_send_packet_confirmed(void **state)
{
    (void) state;
    uint8_t msg[] = {0x11, 0x22, 0x33, 0x44};
    uint8_t cmd1[] = {LL_LORAWAN_SEND_LINK_CHECK | LL_LORAWAN_SEND_CONFIRMED,
            1, 3, 4, 0x11, 0x22, 0x33, 0x44};
    uint8_t rsp_success[] = {0};

    transport_expect(OP_LORAWAN_MSG_SEND, cmd1, sizeof(cmd1), LL_IFC_NACK_OTHER, 0, 0);
    assert_int_equal(-LL_IFC_NACK_OTHER,
            ll_lorawan_send_confirmed(LL_LORAWAN_SEND_LINK_CHECK, 1, msg, sizeof(msg), 3));

    transport_expect(OP_LORAWAN_MSG_SEND, cmd1, sizeof(cmd1), LL_IFC_NACK_BUSY_TRY_AGAIN, 0, 0);
    assert_int_equal(-LL_IFC_NACK_BUSY_TRY_AGAIN,
            ll_lorawan_send_confirmed(LL_LORAWAN_SEND_LINK_CHECK, 1, msg, sizeof(msg), 3));

    transport_expect(OP_LORAWAN_MSG_SEND, cmd1, sizeof(cmd1), 0, rsp_success, sizeof(rsp_success));
    assert_int_equal(0,
            ll_lorawan_send_confirmed(LL_LORAWAN_SEND_LINK_CHECK, 1, msg, sizeof(msg), 3));
}

void test_lorawan_receive(void **state)
{
    (void) state;
    ll_lorawan_rx_t rx;
    uint8_t buffer[256];
    uint8_t msg[] = {0x11, 0x22, 0x33, 0x44};
    uint8_t flags = LL_LORAWAN_RECEIVE_LINK_CHECK | LL_LORAWAN_RECEIVE_MESSAGE | LL_LORAWAN_RECEIVE_ACK;
    uint8_t rsp_success[] = {flags, 3, 10, 2, 180, 11, 5, 4, 0x11, 0x22, 0x33, 0x44};

    transport_expect(OP_LORAWAN_MSG_RECEIVE, 0, 0, 0, rsp_success, sizeof(rsp_success));
    assert_int_equal(0, ll_lorawan_receive(buffer, sizeof(buffer), &rx));
    assert_int_equal(flags, rx.flags);
    assert_int_equal(3, rx.TxNbRetries);
    assert_int_equal(10, rx.DemodMargin);
    assert_int_equal(2, rx.NbGateways);
    assert_int_equal(LL_LORAWAN_RSSI_FROM_PKT(180), rx.RxRssi);
    assert_int_equal(11, rx.RxSnr);
    assert_int_equal(5, rx.RxPort);
    assert_int_equal(sizeof(msg), rx.bytes_received);
    assert_memory_equal(msg, buffer, rx.bytes_received);
}

void test_lorawan_receive_buffer_too_small(void **state)
{
    (void) state;
    ll_lorawan_rx_t rx;
    uint8_t buffer[256];
    uint8_t rsp_success[] = LL_LORAWAN_RECEIVE_MESSAGE, 3, 10, 2, 180, 11, 5, 4, 0x11, 0x22, 0x33, 0x44};

    transport_expect(OP_LORAWAN_MSG_RECEIVE, 0, 0, 0, rsp_success, sizeof(rsp_success));
    assert_int_equal(LL_IFC_ERROR_BUFFER_TOO_SMALL, ll_lorawan_receive(buffer, 2, &rx));
}

void test_lorawan_receive_nack(void **state)
{
    (void) state;
    ll_lorawan_rx_t rx;
    uint8_t buffer[256];

    transport_expect(OP_LORAWAN_MSG_RECEIVE, 0, 0, LL_IFC_NACK_OTHER, 0, 0);
    assert_int_equal(-LL_IFC_NACK_OTHER, ll_lorawan_receive(buffer, 2, &rx));
}

void test_lorawan_receive_payload_too_small(void **state)
{
    (void) state;
    ll_lorawan_rx_t rx;
    uint8_t buffer[256];
    uint8_t rsp_success[] = {LL_LORAWAN_RECEIVE_MESSAGE, 3, 10, 2, 180, 11, 5};

    transport_expect(OP_LORAWAN_MSG_RECEIVE, 0, 0, 0, rsp_success, sizeof(rsp_success));
    assert_int_equal(LL_IFC_ERROR_INCORRECT_MESSAGE_SIZE, ll_lorawan_receive(buffer, 2, &rx));
}

void test_lorawan_receive_payload_size_mismatch(void **state)
{
    (void) state;
    ll_lorawan_rx_t rx;
    uint8_t buffer[256];
    uint8_t rsp_success[] = {LL_LORAWAN_RECEIVE_MESSAGE, 3, 10, 2, 180, 11, 5, 4, 0x11, 0x22, 0x33, 0x44, 0xff};

    transport_expect(OP_LORAWAN_MSG_RECEIVE, 0, 0, 0, rsp_success, sizeof(rsp_success));
    assert_int_equal(LL_IFC_ERROR_INCORRECT_MESSAGE_SIZE, ll_lorawan_receive(buffer, 2, &rx));
}


/* ----------------------- Main -------------------------------------------- */

int main(int argc, char* argv[]) {
    (void) argc;
    (void) argv;

    UnitTest tests[] = {
        unit_test_setup_teardown(test_myself_transport_write, setUp, tearDown),
        unit_test_setup_teardown(test_timestamp_get, setUp, tearDown),
        unit_test_setup_teardown(test_timestamp_get_with_response_error, setUp, tearDown),
        unit_test_setup_teardown(test_timestamp_get_twice_to_check_message_identifier, setUp, tearDown),
        unit_test_setup_teardown(test_timestamp_set, setUp, tearDown),
        unit_test_setup_teardown(test_timestamp_set_with_nack, setUp, tearDown),
        unit_test_setup_teardown(test_nack_with_payload, setUp, tearDown),
        unit_test_setup_teardown(test_timestamp_set_with_invalid_operation, setUp, tearDown),
        unit_test_setup_teardown(test_packet_send_timestamp, setUp, tearDown),

        unit_test_setup_teardown(test_lorawan_activate_over_air, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_activate_personalization, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_param_get_i32_invalid, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_param_get_i32_success, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_param_get_i32_with_payload_too_short, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_param_set_i32_invalid, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_param_set_i32_success, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_param_set_i32_with_payload_too_short, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_send_packet_unconfirmed, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_send_packet_confirmed, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_receive, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_receive_buffer_too_small, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_receive_nack, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_receive_payload_too_small, setUp, tearDown),
        unit_test_setup_teardown(test_lorawan_receive_payload_size_mismatch, setUp, tearDown),
    };

    return run_tests(tests);
}
