/* Copyright 2015 Link Labs LLC.  All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdbool.h>
#include <inttypes.h>
#include "str_utils.h"
#include "embc/log.h"
#include "ll_ifc.h"
#include "ll_ifc_lorawan.h"
#include "ll_ifc_transport_pc.h"


/**
 * \file
 *
 * \brief Host application for LoRaWAN connectivity using Link Labs devices.
 */


#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#if defined(WIN32) || defined (__MINGW32__)
#include <windows.h>
#endif

// Build Date
#define BUILD_DATE (__BUILD_DATE);

// Build Number
#define VERSION_MAJOR                (1) // 0-255
#define VERSION_MINOR                (3) // 0-255
#define VERSION_TAG     (__BUILD_NUMBER) // 0-65535

#define RX_BUF_LEN          (255)
#define USAGE_ERROR         (42)

typedef int (*command_handler)(int argc, char ** argv);

typedef struct command_s {
    const char * name;
    command_handler handler;
} command_t;

typedef struct param_s {
    enum ll_lorawan_param_e id;
    const char * name;
    bool read_only;
} param_t;

static const param_t PARAMS[] = {
    {LL_LORAWAN_PARAM_ADAPTIVE_DATA_RATE_ENABLED, "adaptive-data-rate-enabled", false},
    {LL_LORAWAN_PARAM_NETWORK_ACTIVATION_STATUS, "network-activation-status", true},
    {LL_LORAWAN_PARAM_NETWORK_TYPE, "network-type", true},
    {LL_LORAWAN_PARAM_DEVICE_CLASS, "device-class", true},
    {LL_LORAWAN_PARAM_UPLINK_COUNTER, "uplink-counter", true},
    {LL_LORAWAN_PARAM_DOWNLINK_COUNTER, "downlink-counter", true},
    {(enum ll_lorawan_param_e) 0, 0, 0}
};

static const struct command_s * command_ = 0;


/*
 * Inspired by:
 *   getopt: http://www.gnu.org/software/libc/manual/html_node/Getopt.html
 *   docopt: https://github.com/docopt/docopt.c
 *   yuck: http://www.fresse.org/yuck/
 *
 */
static const char USAGE[] = 
        "Usage:\n"
        "  lorawan_test [options] <command> [args]\n"
        "\n"
        "Options:\n"
        "  -b, --baudrate <baudrate>     Configure baudrate of the tty device\n"
        "  -D, --device <device>         Choose tty device\n"
        "  -h, --help                    Print this help message\n"
        "\n"
        "Commands:\n"
        "\n"
        "  initialize\n"
        "        Configure the module for LoRaWAN.\n"
        "        -h, --hybrid\n"
        "            Use LoRaWAN hybrid mode when applicable (FCC 915 only).\n"
        "\n"
        "  info\n"
        "        Get detailed information about the module.\n"
        "\n"
        "  activate-ota\n"
        "        Activate LoRaWAN over-the-air.\n"
        "        <net_type> enum (public | private)\n"
        "            The network type.\n"
        "        <node_class> enum (A | B | C)\n"
        "            The device node class.\n"
        "        <devEui> hex_array[8]\n"
        "           The device EUI as a hex string.\n"
        "           A value of all zeros uses the preconfigured EUI.\n"
        "        <appEui> hex_array[8]\n"
        "           The application EUI as a hex string.\n"
        "        <appKey> hex_array[16]\n"
        "           The application AES128 key as a hex string.\n"
        "\n"
        "  activate-per\n"
        "        Activate LoRaWAN over-the-air.\n"
        "        <net_type> enum (public | private)\n"
        "           The network type.\n"
        "        <node_class> enum (A | B | C)\n"
        "            The device node class.\n"
        "        <netID> int32\n"
        "           The 24-bit network identifier\n"
        "        <devAddr> int32\n"
        "            The 32-bit device address which must be unique for the network.\n"
        "        <netSKey> hex_array[16]\n"
        "            The network security key as a hex string.\n"
        "        <appSKey> hex_array[16]\n"
        "            The application security key as a hex string.\n"
        "\n"
        "  list\n"
        "        List the available parameters.\n"
        "\n"
        "  get\n"
        "        Get a parameter's value.\n"
        "        <param>\n"
        "            The parameter name to get.\n"
        "\n"
        "  set\n"
        "        Set a parameter's value.\n"
        "        <param>\n"
        "            The parameter name to set.\n"
        "        <value>\n"
        "            The new int32 value for param.\n"
        "\n"
        "  tx\n"
        "        Transmit a message.\n"
        "        -X, --hex\n"
        "            When provided, the message is in hex.  The message is an ASCII\n"
        "            string by default.\n"
        "        -C, --link-check\n"
        "            Perform a link status check\n"
        "        -p, --port <port>\n"
        "            The LoRaWAN application port (1 to 255).  Defaults to 1.\n"
        "        -c, --confirmed\n"
        "            Transmit message with received confirmation requested.\n"
        "        -r, --retries <retries>\n"
        "            The number of retries with --confirmed is specified.\n"
        "            Defaults to 5.\n"
        "        <message>\n"
        "            The string or hex array.\n"
        "\n"
        "  rx\n"
        "        Receive a message.\n"
        "        -X, --hex\n"
        "            When provided, the message is displayed in hex.  The message is\n"
        "            displayed as an ASCII string by default.\n"
        "\n"
        "  irq-get\n"
        "        Get the bit mask of pending interrupt events.\n"
        "\n"
        "  irq-clear\n"
        "        Acknowledge pending interrupt events.\n"
        "        <mask>\n"
        "            The bit mask of interrupt bitbits to acknowledge.\n"
        "            When not specified, default to 0xffffffff to acknowledge all\n"
        "            pending interrupts.\n"
        "\n"
        "  help\n"
        "        Display usage information.\n"
        "\n"
;

//static const char * NETWORK_TYPE_ENUM[] = {"public", "private", 0};
//static const char * DEVICE_CLASS_ENUM[] = {"a", "b", "c", 0};

static int usage(void);

static int command_initialize(int argc, char ** argv);
static int command_info(int argc, char ** argv);
static int command_activate_ota(int argc, char ** argv);
static int command_activate_per(int argc, char ** argv);
static int command_list(int argc, char ** argv);
static int command_get(int argc, char ** argv);
static int command_set(int argc, char ** argv);
static int command_tx(int argc, char ** argv);
static int command_rx(int argc, char ** argv);
static int command_irq_get(int argc, char ** argv);
static int command_irq_clear(int argc, char ** argv);
static int command_help(int argc, char ** argv);


uint32_t log_timestamp_us( void ) {
    return 0;
}

void log_printf(const char *format, ...)
{
    va_list arg;
    va_start(arg, format);
    vprintf(format, arg);
    va_end(arg);
}

void dbc_assert(char const *file, unsigned line, const char * msg) {
    (void) file;
    (void) line;
    (void) msg;
}

int argument_error(const char * format, ...)
{
    va_list arg;
    va_start(arg, format);
    vprintf(format, arg);
    va_end(arg);
    printf("\n\n");
    return usage();
}

static int print_eui64(void)
{
    uint64_t uuid = 0;
    int8_t ret;

    ret = ll_unique_id_get(&uuid);
    if (ret < 0)
    {
        LOGF_ERROR("ll_unique_id_get failed: %d", ret);
        return -ret;
    }
    else
    {
        printf("Link Labs EUI-64: %016" PRIx64 "\n", uuid);
        printf("Link Labs UUID: $301$0-0-0-%01x", (uint32_t)(uuid >> 32));
        printf("%08x\n", (uint32_t) (uuid & 0xFFFFFFFFF));
        return 0;
    }
}

static int rc_to_error(int8_t rc)
{
    if (rc < 0)
    {
        return -rc;
    }
    return LL_IFC_NACK_PAYLOAD_LEN_OOR;
}

static int print_firmware_type(void)
{
    ll_firmware_type_t t;
    int8_t ret;

    ret = ll_firmware_type_get(&t);
    if (ret != FIRMWARE_TYPE_LEN)
    {
        LOGF_ERROR("ll_firmware_type_get failed: %d", ret);
        return rc_to_error(ret);
    }
    else
    {
        printf("Link Labs Firmware Type: %04d.%04d\n", t.cpu_code, t.functionality_code);
        return 0;
    }
}

static int print_hardware_type(void)
{
    ll_hardware_type_t t;
    int8_t ret;

    ret = ll_hardware_type_get(&t);
    if (ret != 1)
    {
        LOGF_ERROR("ll_hardware_type_get failed: %d", ret);
        return rc_to_error(ret);
    }
    else
    {
        printf("Link Labs Hardware Type: %s\n", ll_hardware_type_string(t));
        return 0;
    }
}


static const command_t COMMANDS[] = {
    {"initialize",      command_initialize},
    {"info",            command_info},
    {"activate-ota",    command_activate_ota},
    {"activate-per",    command_activate_per},
    {"list",            command_list},
    {"get",             command_get},
    {"set",             command_set},
    {"tx",              command_tx},
    {"rx",              command_rx},
    {"irq-get",         command_irq_get},
    {"irq-clear",       command_irq_clear},
    {"help",            command_help},
    {0,                 0}
};

static int usage(void)
{
    if (0 == command_)
    {
        printf("%s", USAGE);
    }
    else
    {
        const char * b;
        const char * buf_end = USAGE + sizeof(USAGE);
        char buf[256];
        snprintf(buf, sizeof(buf), "\n  %s", command_[0].name);
        b = strstr(USAGE, buf);
        if (!b) {
            printf("%s", USAGE);
            return USAGE_ERROR;
        }
        b += 3;
        if (0 != command_[1].name) {
            const char * be;
            snprintf(buf, sizeof(buf), "\n  %s", command_[1].name);
            be = strstr(USAGE, buf);
            if (be)
            {
                buf_end = be;
            }
        }
        printf("usage:\n  lorawan_test [options] ");
        snprintf(buf, sizeof(buf), "%%.%ds", (int) (buf_end - b));
        printf(buf, b);
    }
    return USAGE_ERROR;
}

#define ARG_ADVANCE()  --argc; if (argc < 0) {return usage(); }; ++argv
#define ARG_REQUIRE(source) if (argc < 1) {printf("Missing %s argument\n", source); return usage(); }
#define ARG_ADVANCE_AND_REQUIRE(source) ARG_ADVANCE(); ARG_REQUIRE(source)

int command_info(int argc, char ** argv)
{
    (void) argv;
    int rv = 0;

    if (argc != 0)
    {
        return argument_error("Received %d arguments when 0 where expected", argc);
    }
    rv |= print_firmware_type() ? 1 : 0;
    rv |= print_hardware_type() ? 1 : 0;
    rv |= print_eui64() ? 1 : 0;
    return rv;
}

int command_initialize(int argc, char ** argv)
{
    (void) argv;
    int32_t rv;
    uint32_t irq_flags = 0;
    ll_mac_type_t mac_type = LORAWAN;

    while (argc)
    {
        if (*argv[0] != '-')
        {
            break;
        }
        if ((0 == strcmp(*argv, "--hybrid")) || (0 == strcmp(*argv, "-h")))
        {
            mac_type = LORAWAN_HYBRID;
        }
        else
        {
            return argument_error("Unsupported argument %d: %s", argc, *argv);
        }
        ARG_ADVANCE();
    }

    if (argc != 0)
    {
        return argument_error("Received %d arguments when 0 where expected", argc);
    }

    rv = ll_mac_mode_set(mac_type);
    if (rv < 0)
    {
        LOGF_ERROR("ll_mac_mode_set returned %s", rv);
        return -rv;
    }

    rv = ll_irq_flags(0xffffffff, &irq_flags);
    if (rv < 0)
    {
        LOGF_ERROR("ll_irq_flags returned %s", rv);
        return -rv;
    }

    return 0;
}

static const char * const NETWORK_TYPE_TABLE[] = {"PUBLIC", "PRIVATE", NULL};
static const char * const NETWORK_CLASS_TABLE[] = {"A", "B", "C", NULL};

int command_activate_ota(int argc, char ** argv)
{
    int network_type = 0;
    int network_class = 0;
    size_t sz;

    uint8_t devEui[8];
    uint8_t appEui[8];
    uint8_t appKey[16];

    memset(devEui, 0, sizeof(devEui));
    memset(appEui, 0, sizeof(appEui));
    memset(appKey, 0, sizeof(appKey));

    if (argc != 5)
    {
        return argument_error("Received %d arguments but 5 where expected", argc);
    }

    safestr_toupper(argv[0]);
    if (safestr_toIndex(argv[0], NETWORK_TYPE_TABLE, &network_type))
    {
        return argument_error("Invalid network_type: %s", argv[0]);
    }

    safestr_toupper(argv[1]);
    if (safestr_toIndex(argv[1], NETWORK_CLASS_TABLE, &network_class))
    {
        return argument_error("Invalid network_type: %s", argv[1]);
    }

    if (strcmp(argv[2], "0") == 0)
    {
        // use default device EUI.
        sz = 1;
    }
    else if (hex_string_to_bytes(argv[2], devEui, sizeof(devEui), &sz))
    {
        return argument_error("Invalid devEui: %s", argv[2]);
    }
    if ((sz == sizeof(devEui)) || ((sz == 1) && (devEui[0] == 0)))
    {
        // valid data
    }
    else
    {
        return argument_error("Invalid devEui length: %s", argv[2]);
    }

    if (hex_string_to_bytes(argv[3], appEui, sizeof(appEui), &sz)
            || (sz != sizeof(appEui)))
    {
        return argument_error("Invalid appEui: %s", argv[3]);
    }

    if (hex_string_to_bytes(argv[4], appKey, sizeof(appKey), &sz)
            || (sz != sizeof(appKey)))
    {
        return argument_error("Invalid appEui: %s", argv[4]);
    }

    return ll_lorawan_activate_over_the_air(
            (enum ll_lorawan_network_type_e) network_type,
            (enum ll_lorawan_device_class_e) network_class,
            devEui, appEui, appKey);
}

int command_activate_per(int argc, char ** argv)
{
    int network_type = 0;
    int network_class = 0;
    size_t sz;
    uint32_t netID = 0;
    uint32_t devAddr = 0;

    uint8_t netSKey[16];
    uint8_t appSKey[16];

    memset(netSKey, 0, sizeof(netSKey));
    memset(appSKey, 0, sizeof(appSKey));

    if (argc != 6)
    {
        return argument_error("Received %d arguments when 6 where expected", argc);
    }

    safestr_toupper(argv[0]);
    if (safestr_toIndex(argv[0], NETWORK_TYPE_TABLE, &network_type))
    {
        return argument_error("Invalid network_type: %s", argv[0]);
    }

    safestr_toupper(argv[1]);
    if (safestr_toIndex(argv[1], NETWORK_CLASS_TABLE, &network_class))
    {
        return argument_error("Invalid network_type: %s", argv[1]);
    }

    if (str_to_uint32(argv[2], &netID))
    {
        return argument_error("Invalid netID: %s", argv[2]);
    }

    if (str_to_uint32(argv[3], &devAddr))
    {
        return argument_error("Invalid devAddr: %s", argv[3]);
    }

    if (hex_string_to_bytes(argv[4], netSKey, sizeof(netSKey), &sz)
            || (sz != sizeof(netSKey)))
    {
        return argument_error("Invalid netSKey: %s", argv[4]);
    }

    if (hex_string_to_bytes(argv[5], appSKey, sizeof(appSKey), &sz)
            || (sz != sizeof(appSKey)))
    {
        return argument_error("Invalid appSKey: %s", argv[5]);
    }

    return ll_lorawan_activate_personalization(
            (enum ll_lorawan_network_type_e) network_type,
            (enum ll_lorawan_device_class_e) network_class,
            netID, devAddr, netSKey, appSKey);
}

int command_list(int argc, char ** argv)
{
    (void) argv;
    const param_t * p = PARAMS;
    if (argc != 0)
    {
        return argument_error("Received %d arguments when 0 where expected", argc);
    }
    while (p && p->name)
    {
        printf("%-30s %s\n", p->name, p->read_only ? "read-only": "");
        ++p;
    }
    return 0;
}

static param_t const * find_param(char * name)
{
    const param_t * p = PARAMS;
    cmd_reformat(name);

    while (1)
    {
        if (!p && !p->name)
        {
            argument_error("Invalid parameter: %s\n", name);
            return 0;
        }
        if (strcmp(p->name, name) == 0)
        {
            return p;
        }
        ++p;
    }
}

int command_get(int argc, char ** argv)
{
    const param_t * p = 0;
    int32_t value = 0;
    int32_t rv;

    if (argc != 1)
    {
        return argument_error("Received %d arguments when 1 expected", argc);
    }

    p = find_param(*argv);
    if (!p)
    {
        return USAGE_ERROR;
    }

    rv = ll_lorawan_param_get_i32(p->id, &value);
    if (rv == 0)
    {
        printf("%s = %d (0x%08x)\n", p->name, value, value);
    }

    return rv;
}

int command_set(int argc, char ** argv)
{
    const param_t * p = 0;
    int32_t value = 0;

    if (argc != 2)
    {
        return argument_error("Received %d arguments when 2 expected", argc);
    }

    p = find_param(*argv);
    if (!p)
    {
        return USAGE_ERROR;
    }

    if (!str_to_int32(argv[1], &value))
    {
        return argument_error("Invalid integer value %s", argv[1]);
    }

    return ll_lorawan_param_set_i32(p->id, value);
}

int command_tx(int argc, char ** argv)
{
    bool ishex = false;
    uint8_t port = 1;
    bool confirmed = false;
    uint8_t retries = 5;
    size_t msg_len = 0;
    uint8_t flags = 0;

    uint8_t msg[256];

    while (argc)
    {
        if (*argv[0] != '-')
        {
            break;
        }
        if ((0 == strcmp(*argv, "--hex")) || (0 == strcmp(*argv, "-X")))
        {
            ishex = true;
        }
        else if ((0 == strcmp(*argv, "--link-check")) || (0 == strcmp(*argv, "-C")))
        {
            flags |= LL_LORAWAN_SEND_LINK_CHECK;
        }
        else if ((0 == strcmp(*argv, "--port")) || (0 == strcmp(*argv, "-p")))
        {
            ARG_ADVANCE_AND_REQUIRE("port");
            if (str_to_uint8(*argv, &port) || (port == 0))
            {
                return argument_error("Invalid port value: %s", *argv);
            }
        }
        else if ((0 == strcmp(*argv, "--confirmed")) || (0 == strcmp(*argv, "-c")))
        {
            confirmed = true;
        }
        else if ((0 == strcmp(*argv, "--retries")) || (0 == strcmp(*argv, "-r")))
        {
            ARG_ADVANCE_AND_REQUIRE("retries");
            if (str_to_uint8(*argv, &retries))
            {
                return argument_error("Invalid retries value: %s", *argv);
            }
        }
        else
        {
            return argument_error("Unsupported argument %d: %s", argc, *argv);
        }
        ARG_ADVANCE();
    }

    ARG_REQUIRE("message");

    if (ishex)
    {
        if (hex_string_to_bytes(*argv, msg, sizeof(msg), &msg_len))
        {
            return argument_error("Invalid hex string: %s", *argv);
        }
    }
    else
    {
        msg_len = strlen(argv[0]);
        if (msg_len > 255)
        {
            return argument_error("String too long: %d", msg_len);
        }
        memcpy(msg, argv[0], msg_len);
    }

    if (confirmed)
    {
        return ll_lorawan_send_confirmed(flags, port, msg, (uint8_t) msg_len, retries);
    }
    else
    {
        return ll_lorawan_send_unconfirmed(flags, port, msg, (uint8_t) msg_len);
    }
}

int command_rx(int argc, char ** argv)
{
    bool ishex = false;
    uint8_t msg[257];
    ll_lorawan_rx_t rx;
    int32_t rv;
    uint32_t irq_flags_clear = IRQ_FLAGS_RX_DONE |
                               IRQ_FLAGS_TX_DONE |
                               IRQ_FLAGS_TX_ERROR;
    uint32_t irq_flags_read = 0;

    memset(msg, 0, sizeof(msg));
    memset(&rx, 0, sizeof(rx));

    while (argc)
    {
        if (*argv[0] != '-')
        {
            break;
        }
        if ((0 == strcmp(*argv, "--hex")) || (0 == strcmp(*argv, "-X")))
        {
            ishex = true;
        }
        else
        {
            return argument_error("Unsupported argument %d: %s", argc, *argv);
        }
        ARG_ADVANCE();
    }
    if (argc > 0)
    {
        return argument_error("Unsupported argument %d: %s", argc, *argv);
    }

    rv = ll_irq_flags(irq_flags_clear, &irq_flags_read);
    if (rv < 0)
    {
        LOGF_WARN("ll_irq_flags returned %d", rv);
    }

    rv = ll_lorawan_receive(msg, sizeof(msg) - 1, &rx);
    if (rv < 0)
    {
        return -rv;
    }

    printf("  flags =");
    if (rx.flags & LL_LORAWAN_RECEIVE_LINK_CHECK)
    {
        printf(" LINK_CHECK");
    }
    if (rx.flags & LL_LORAWAN_RECEIVE_ACK)
    {
        printf(" ACK");
    }
    if (rx.flags & LL_LORAWAN_RECEIVE_MESSAGE)
    {
        printf(" RX");
    }
    if (rx.flags & LL_LORAWAN_RECEIVE_MESSAGE_MULTICAST)
    {
        printf(" MULTICAST");
    }
    printf("\n");

    printf("  TxAckReceived = %d\n", rx.TxAckReceived);
    printf("  TxNbRetries = %d\n", rx.TxNbRetries);
    printf("  bytes_received = %d\n", rx.bytes_received);
    printf("  RxPort = %d\n", rx.RxPort);
    printf("  RxRssi = %d\n", rx.RxRssi);
    printf("  RxSnr = %d\n", rx.RxSnr);
    printf("  DemodMargin = %d\n", rx.DemodMargin);
    printf("  NbGateways = %d\n", rx.NbGateways);

    if (rx.flags & LL_LORAWAN_RECEIVE_MESSAGE)
    {
        printf("message:\n");
        if (ishex)
        {
            size_t idx;
            for (idx = 0; idx < rx.bytes_received; ++idx)
            {
                printf("%s%02x", (idx > 0) ? " " : "", msg[idx]);
            }
        }
        else
        {
            printf("%s", msg);
        }
        printf("\n");
    }
    return 0;
}

static void print_irq_flags(uint32_t flags)
{
    printf("IRQ:");

    if (flags == 0)
    {
        printf(" __none__");
    }

    // Most significant bit to least significant bit
    if(IRQ_FLAGS_ASSERT & flags)
    {
        printf(" ASSERT");
    }
    if(IRQ_FLAGS_APP_TOKEN_ERROR & flags)
    {
        printf(" APP_TOKEN_ERROR");
    }
    if(IRQ_FLAGS_CRYPTO_ERROR & flags)
    {
        printf(" CRYPTO_ERROR");
    }
    if(IRQ_FLAGS_DOWNLINK_REQUEST_ACK & flags)
    {
        printf(" DOWNLINK_REQUEST_ACK");
    }
    if(IRQ_FLAGS_INITIALIZATION_COMPLETE & flags)
    {
        printf("[IRQ_FLAGS_INITIALIZATION_COMPLETE]");
    }
    if(IRQ_FLAGS_APP_TOKEN_CONFIRMED & flags)
    {
        printf(" APP_TOKEN_CONFIRMED");
    }
    if(IRQ_FLAGS_CRYPTO_ESTABLISHED & flags)
    {
        printf(" CRYPTO_ESTABLISHED");
    }
    if(IRQ_FLAGS_DISCONNECTED & flags)
    {
        printf(" DISCONNECTED");
    }
    if(IRQ_FLAGS_CONNECTED & flags)
    {
        printf(" CONNECTED");
    }
    if(IRQ_FLAGS_RX_DONE & flags)
    {
        printf(" RX_DONE");
    }
    if(IRQ_FLAGS_TX_ERROR & flags)
    {
        printf(" TX_ERROR");
    }
    if(IRQ_FLAGS_TX_DONE & flags)
    {
        printf(" TX_QUEUE_EMPTY");
    }
    if(IRQ_FLAGS_RESET & flags)
    {
        printf(" RESET");
    }
    if(IRQ_FLAGS_WDOG_RESET & flags)
    {
        printf(" WDOG_RESET");
    }
    printf("\n");
}

static int command_irq_get(int argc, char ** argv)
{
    (void) argc;
    (void) argv;
    int8_t i8_ret;
    uint32_t irq_flags = 0;
    i8_ret = ll_irq_flags(0, &irq_flags);
    if (i8_ret < 0)
    {
        printf("Could not get IRQ: %d\n", i8_ret);
        return (int) -i8_ret;
    }
    print_irq_flags(irq_flags);
    return 0;
}

static int command_irq_clear(int argc, char ** argv)
{
    uint32_t irq_flags_clear = 0xffffffff;
    int8_t i8_ret;
    uint32_t irq_flags_read = 0;

    if (argc == 0)
    {
        // no worries!
    }
    else if (argc == 1)
    {
        if (str_to_uint32(argv[0], &irq_flags_clear))
        {
            return argument_error("Invalid flags value: %s", argv[0]);
        }
    }

    i8_ret = ll_irq_flags(irq_flags_clear, &irq_flags_read);
    if (i8_ret < 0)
    {
        return (int) -i8_ret;
    }

    return 0;
}

struct command_s const * find_command(char const * name)
{
    struct command_s const * cmd_instance = COMMANDS;
    char buf[256];

    snprintf(buf, sizeof(buf), "%s", name);
    cmd_reformat(buf);

    while (1)
    {
        if (!cmd_instance || !cmd_instance->name)
        {
            argument_error("Unknown command: %s", name);
            return 0;
        }
        if (strcmp(cmd_instance->name, buf) == 0) {
            return cmd_instance;
        }
        ++cmd_instance;
    }
    return 0;
}

int command_help(int argc, char ** argv)
{
    command_ = 0;
    if (argc == 0)
    {
        command_ = 0;
    }
    else if (argc == 1)
    {
        command_ = find_command(argv[0]);
        if (0 == command_)
        {
            return USAGE_ERROR; // error message already displayed
        }
    }
    usage();
    return 0;
}

int main (int argc, char ** argv)
{
    int rc = 0;
    int32_t baudrate = 115200;
    const char * tty_dev_name = 0;
    ARG_ADVANCE(); // consume filename.
    
    while (argc > 0)
    {
        if (*argv[0] != '-')
        {
            break;
        }
        if ((0 == strcmp(*argv, "--baudrate")) || (0 == strcmp(*argv, "-b")))
        {
            ARG_ADVANCE_AND_REQUIRE("baudrate");
            baudrate = strtol(*argv, 0, 10);
        }
        else if ((0 == strcmp(*argv, "--device")) || (0 == strcmp(*argv, "-D")))
        {
            ARG_ADVANCE_AND_REQUIRE("device");
            tty_dev_name = *argv;
        }
        else if ((0 == strcmp(*argv, "--help")) || (0 == strcmp(*argv, "-h")))
        {
            return command_help(0, 0);
        }
        else
        {
            return argument_error("Unsupported argument %d: %s", argc, *argv);
        }
        ARG_ADVANCE();
    }
    
    if (argc <= 0)
    {
        return argument_error("No command provided");
    }
    
    command_ = find_command(argv[0]);
    if (!command_)
    {
        return USAGE_ERROR; // error message already displayed
    }
    ARG_ADVANCE();

    if (command_->handler != command_help)
    {
        LOGF_INFO("ll_tty_open(%s, %d)", tty_dev_name, baudrate);
        rc = ll_tty_open(tty_dev_name, baudrate);
        if (rc)
        {
            return argument_error("ll_tty_open failed with %d", rc);
        }
    }
    
    LOGF_INFO("Command \"%s\" with %d arguments", command_->name, argc);
    rc = command_->handler(argc, argv);
    if (rc)
    {
        if (rc > 0)
        {
            rc = -rc;
        }
        printf("Command returned error %d: %s\n", rc, ll_return_code_name(rc));
        rc = (rc > 0) ? -rc : rc;  // make negative
        printf("%s\n", ll_return_code_description(rc));
        rc = -rc;  // make positive
    }

    ll_tty_close();

    return rc;
}
